package ch.lanter.vajanoid;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.RenderingHints;

/**
 * This class contains the state of a ball. This includes the position, speed,
 * direction and image
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class Ball implements Positionable, Eventable {

	private static final long serialVersionUID = 201001130140L;

	private Model model; // only for ball events

	// Position of the top left pixel
	public float x,y;
	public int width, height;

	// Speed in pixel per second
	public float vx, vy; // in x and y direction
	protected float velocity; // in total

	protected boolean bind; // true if the ball should bind to the pad
	protected boolean bound; // true if the ball is bound to the pad

	private boolean invincible; // true if the ball is invicible (golden)

	public Ball(Model model) {
		this.model = model;
		this.x = 20f;
		this.y = 240f;
		this.velocity = 256;
		this.vx = 0;
		this.vy = velocity;
		this.width = 12;
		this.height = 12;
		this.bind = false;
		this.bound = false;
		this.invincible = false;
	}

	public Ball(Ball ball) { // semi-deep copy
		this.model = ball.model;
		this.x = ball.x;
		this.y = ball.y;
		this.velocity = ball.velocity;
		this.vx = ball.vx;
		this.vy = ball.vy;
		this.width = ball.width;
		this.height = ball.height;
		this.bind = ball.bind;
		this.bound = ball.bound;
		this.invincible = ball.invincible;
	}

	/**
	 * Adjust your position according to the passed time dt
	 * @param dt - time passed since the last call
	 */
	public void step(long dt) {
		if (!bound) {
			x = x + vx * dt/1000;
			y = y + vy * dt/1000;
		}
	}

	/**
	 * Switch horizontal direction of the ball
	 */
	public void switchVx() {
		vx = -vx;
	}

	/**
	 * Switch vertical direction of the ball
	 */
	public void switchVy() {
		vy = -vy;
	}

	public void setInvincible(boolean b) {
		invincible = b;
	}

	public boolean isInvincible() {
		return invincible;
	}

	// delegate
	public void addScore(int add) {
		model.addScore(add);
	}

	public void launchEvent(BallEvent event) {
		model.notifyObservers(event);
	}

	@Override
	public void paint(Graphics g) {
		Image image = null;
		if (invincible) {
			image = ImageCenter.getInstance().imgBall2; // inviincible (golden) ball
		} else {
			image = ImageCenter.getInstance().imgBall; // normal (iron) ball
		}
		Graphics2D g2 = (Graphics2D) g.create();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.drawImage(image,getX(),getY(), width, height, null);
		g2.dispose();
	}

	public Rectangle getBounds() {
		return new Rectangle(getX(),getY(),width,height);
	}

	public int getX() {
		return Math.round(x);
	}

	public int getY() {
		return Math.round(y);
	}

	public int getWidth() {
		return Math.round(width);
	}

	public int getHeight() {
		return Math.round(height);
	}
}
