package ch.lanter.vajanoid;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.util.Random;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.Timer;
import javax.swing.border.EtchedBorder;

import ch.lanter.vajanoid.block.Block;
import ch.lanter.vajanoid.gui.PnlSidebar;
import ch.lanter.vajanoid.gui.View;
import ch.lanter.vajanoid.net.PnlRank;
import ch.lanter.vajanoid.net.Rank;
import ch.lanter.vajanoid.net.Ranking;

/**
 * The controller instantiates the model and the view. Furtermore it setups a
 * timer, firing a ActionEvent every 10 ms which makes the model update itself
 * by calling model.step(dt). Then the controller checks the state of the model
 * by calling the step(dt) and correct it (e.g. remove lost balls, shots).
 * Moreover the controller checks whether the level has been won or even the
 * whole game or if it is game over and calls adequad methods. The handling of
 * items is predominantly outsourced to the class ItemController.
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class Controller {

	public static boolean printWarnings = false;

	private Timer timer; // timer to launch updates of the model and repaint the view
	private long currTime; // time of last update
	private boolean end; // true if the game is finished (game over, game won, otherwise)

	private Model model;
	private View view;

	private ItemController itemController; // outsource item handling to the class ItemController

	public Controller(JPanel pnlView) {
		model = new Model(); // State of the game
		view = new View(model); // The component with the blocks, balls and the whole action

		itemController = new ItemController(model);
		model.addObserver(itemController);

		timer = new Timer(20, new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				long t0 = System.nanoTime();
				long dt = (System.nanoTime()-currTime)/1000000;
				currTime = System.nanoTime();
				
				model.step(dt<=30 ?dt :30); // let the model update its state
				step(); // controll the state and update all components 
				
				long te = (System.nanoTime()-t0)/1000;
				if (printWarnings && dt>30) System.err.println("Warning: too long dt elapsed: "+dt);
				if (printWarnings && te>5000)  System.err.println("Warning: too long step time elapsed: "+te+"mms");
			}
		}) {
			private static final long serialVersionUID = 201001130141L;
			@Override public void start() { // if the timer starts it has to capture the current time
				currTime = System.nanoTime();
				super.start();
			}
		}; // Looks terrible, I know... ;)

		// Sidebar to the right
		PnlSidebar pnlSidebar = new PnlSidebar(model,new CommandListener());

		// Mouselistener to launch shots if the mouse is clicked
		ShotListener shotListener = new ShotListener();
		view.addMouseListener(shotListener);
		pnlView.addMouseListener(shotListener);

		// set up the Container for the game
		final JPanel pnl = new JPanel();
		pnl.setLayout(new BoxLayout(pnl,BoxLayout.X_AXIS));
		pnl.setOpaque(false);
		pnl.add(view);
		pnl.add(pnlSidebar);
		pnl.setBorder(new EtchedBorder(EtchedBorder.RAISED));

		// to capture mousemovements out of the view component
		pnlView.addMouseMotionListener(new MouseMotionAdapter() {
			@Override public void mouseMoved(MouseEvent e) {
				model.getPad().moveTo(e.getX()-pnl.getX());
			}

			@Override public void mouseDragged(MouseEvent e) {
				model.getPad().moveTo(e.getX()-pnl.getX());
			}
		});

		pnlView.setLayout(new BoxLayout(pnlView,BoxLayout.X_AXIS));
		pnlView.add(Box.createHorizontalGlue());
		pnlView.add(pnl);
		pnlView.add(Box.createHorizontalGlue());
	}

	/**
	 * Restart game
	 */
	public void restart() {
		end = false;
		view.unload();
		model.restart();
		start();
	}

	/**
	 * start first level when ready
	 */
	public void start() {
		model.nextLevel(1);
		timer.start();
	}

	/**
	 * switch pause mode
	 */
	private void switchPause() {
		if (end) return;
		if (timer.isRunning()) {
			setPause(true);
		} else {
			setPause(false);
		}
	}

	/**
	 * pause or resume the game
	 */
	public void setPause(boolean doPause) {
		if (end) return;
		if (doPause) {
			timer.stop();
			model.getLevel().pause();
			view.setStringInFront("Pause","press F3");
		} else {
			view.removeStringInFront();
			timer.start();
			model.getLevel().resume();
		}
	}

	public boolean isRunning() {
		return timer.isRunning();
	}

	/**
	 * After a ball is lost, the player continus with the next ball, if available
	 */
	private void nextBall() {
		timer.stop();
		model.reset(); // new ball, new pad

		for (Ball ball:model.getBalls())
			model.getPad().padBall(ball);
		timer.start();
	}

	/**
	 * Forward in the LevelChain by 1
	 */
	private void nextLevel() {
		nextLevel(1);
	}

	/**
	 * Forward in the LevelChain by 'jumps'
	 */
	private void nextLevel(int jumps) {
		if (model.hasNextLevel()) {
			timer.stop();
			view.unload();
			model.reset();
			model.nextLevel(jumps);

			for (Ball ball:model.getBalls()) model.getPad().padBall(ball);
			new Thread(new Runnable() {
				public void run() {
					try {
//						Thread.sleep(1000);
					} catch (Exception e) {
						e.printStackTrace();
					} finally {
						timer.start();
					}
				}
			}).start();
		} else {
			gameWon();
		}
	}

	/**
	 * Every 10ms update all objects
	 */
	public void step() { // some of these instructions might be moved into the model
		updateBall();
		updateBlocks();
		updateItems();
		updateShots();
		view.repaint();
	}

	/**
	 * Update the Shot objects, check whether they hit a block or ran out of the panel
	 */
	private void updateShots() {
		Shot[] shots = model.getShots();
		Block[] blocks = model.getLevel().getBlocks();
		for (Shot shot:shots) {
			if (shot.getY()<0-shot.width // out of panel
					|| shot.isCatched()) {
				model.removeShot(shot);
			} else {
				for (Block block:blocks) {
					if (block!=null && block.intersects(shot)) { // contact with a block
						block.interract(shot);
					}
				}
			}
		}
	}

	/**
	 * Update the Item objects, check whether they have been catched by the
	 * panel or ran out of the panel
	 */
	private void updateItems() {
		Item[] items = model.getItems();
		Pad pad = model.getPad();
		for (Item i:items) {
			if (i!=null && i.intersects(pad)) { // contact with the pad => catched
				itemController.capture(i);
			}
			if (i!=null && (i.isCatched() || i.getY()>model.getWidth())) { // out of the panel
				model.removeItem(i); // remove item
			}
		}
	}

	/**
	 * Update Block objects, wheter they have been hit by a ball or wheter none is alive any more
	 */
	private void updateBlocks() {
		Block[] blocks = model.getLevel().getBlocks();
		for (Ball ball:model.getBalls()) {
			for (Block b:blocks) {
				if (b!=null && b.intersects(ball)) {
					b.interract(ball);
				}
			}
		}
		if (model.getLevel().isWon()) { // all blocks death (not equal to all are disapeared)
			nextLevel();
		}
	}

	/**
	 * Update the balls. If their position exceeds the game area, make them
	 * switch the direction or if they're lost, remove them
	 */
	private void updateBall() {
		for (Ball ball:model.getBalls()) {
			int width = model.getWidth();
			if (ball.x<0) {
				ball.x = Math.abs(ball.x); // ball pos must be valid, because it might happan, that it doesn't get out of invalid pos only by vx,vy
				ball.switchVx();
			} else if (ball.x+ball.width > width) {
				ball.x = 2*width - ball.x - 2*ball.width;
				ball.switchVx();
			}
			if (ball.y<0) {
				ball.y = Math.abs(ball.y);
				ball.switchVy();
			} else if (Math.floor(ball.y)-ball.height>=model.getHeight()) {
				model.removeBall(ball);
			} else if (Math.floor(ball.y)+ball.height>model.getPad().getPosY()+4) {  // +4 because the pad starts 4 pixel after its y position
				// Ball below the pad
				model.getPad().padBall(ball);
			}
		}
		if (model.getBallCount()<=0) { // all balls lost
			model.decreaseLives();
			if (model.getLives()<0) { // all lives lost
				gameOver();
			} else {
				nextBall();
			}
		}
	}

	/**
	 * Game over
	 */
	private void gameOver() {
		System.out.println("Game Over");
		end = true;
		model.getLevel().stop();
		timer.stop();
		view.setStringInFront("Game Over","press F2");
		ranking();
	}

	/**
	 * The game ist won
	 */
	private void gameWon() {
		System.out.println("Game Won");
		end = true;
		model.getLevel().stop();
		timer.stop();
		view.setStringInFront("Game Won","press F2");
		ranking();
	}

	/**
	 * Download ranking, show it and be ready to upload the score
	 */
	private void ranking() {
		new Thread() {
			public void run() {
				Ranking ranking = new Ranking();
				int score = model.hasCheated()?0:model.getScore();
				int level = model.getLevelNumber();
				Rank rank = new Rank("[Player]",score,level,"Now");
				PnlRank pnlRank = new PnlRank(ranking, rank);
				pnlRank.setSize(view.getSize());
				view.add(pnlRank,0);
				view.validate();
				pnlRank.appear();
			}
		}.start();
	}

	/**
	 * Jump to the level described in the 'cmd'
	 */
	private void jumpToLevel(String cmd) {
		String opcode = cmd.substring(Command.LEVEL.length()).trim();
		int jumps = model.getJumpsToLevel(opcode);
		nextLevel(jumps);
	}

	/**
	 * MouseListener to setup new shots when clicking on the mouse while the pad
	 * is able to shot
	 */
	private class ShotListener extends MouseAdapter {
		public void mouseReleased(MouseEvent event) {
			if (event.getButton()!=MouseEvent.BUTTON1) return; // only leftclicks
			if (!isRunning()) return; // only if running
			fire();
		}
	}

	/**
	 * Fire as many shots as you have canons on the pad
	 */
	private void fire() {
		Pad pad = model.getPad();
		pad.releaseBoundedBalls();
		int shots = model.getShotCount();
		int width = model.getPad().getPadWidth();
		int yy = pad.getPosY();
		for (int i=0;i<shots;i++) {
			int xx = pad.getPosX()+(i+1)*width/(shots+1);
			Shot shot = new Shot(model,xx,yy);
			model.addShot(shot);
		}
	}

	/**
	 * Command Listener to listen to all commands arriving from the buttons and
	 * the cheat Textfield
	 */
	private class CommandListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			String command = event.getActionCommand().toLowerCase().trim();
			if (command.equals(Command.START)) {
				restart();

			} else if (command.equals(Command.FIRE)) {
				fire();

			} else if (command.equals(Command.PAUSE)) {
				switchPause();

			} else if (command.equals(Command.UNLOCK)) {
				Random rand = new Random();
				for (Ball ball:model.getBalls()) {
					double angle = rand.nextDouble()*2*Math.PI;
					ball.vx = (float) Math.cos(angle)*ball.velocity;
					ball.vy = (float) Math.sin(angle)*ball.velocity;
				}

			} else if (command.equals(Command.NEXT)) {
				nextLevel();

			} else if (command.equals(Command.LIVES)) {
				for (int i=model.getLives();i<100;i++) model.increaseLives();

			} else if (command.equals(Command.ZERO)) {
				for (int i=model.getLives();i>0;i--) model.decreaseLives();

			} else if (command.startsWith(Command.ITEMS)) {
				String opcode = command.substring(Command.ITEMS.length()).trim();
				int improp = 1;
				try { improp = Integer.parseInt(opcode); } catch (NumberFormatException e) {}
				itemController.setItemImprobability(improp);

			} else if (command.equals(Command.SHOT)) {
				model.increaseShots();

			} else if (command.equals(Command.GOLD)) {
				for (Ball ball:model.getBalls()) ball.setInvincible(true);

			} else if (command.equals(Command.OVER)) {
				gameOver();

			} else if (command.equals(Command.WINNER)) {
				gameWon();
			} else if (command.equals(Command.LANTERSOFT)) {
				itemController.lantersoftCheat();

			} else if (command.startsWith(Command.LEVEL)) {
				if (!end) jumpToLevel(command);

			} else {
				System.out.println("Unknown command: "+command);
				return;
			}
			if (Command.isCheat(command)) {
				model.setCheated(true);
			}

			System.out.println("Command: "+command);
			view.repaint();
		}
	}
}
