package ch.lanter.vajanoid;

import java.util.Observable;
import java.util.Observer;
import java.util.Random;

import ch.lanter.vajanoid.BallEvent.TYPE;

/**
 * ItemController manages the item handling. It calls adequade metheds if an
 * item has been catched and provides some functionality for launching items
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class ItemController implements Observer {

	private static int defaultItemImpropability = 3; // an item is released every 3rd time a block is destoyed
	private static int ballMax = 500; // limit of balls (might be higher on fast computers)
	
	private Model model;
	
	private int itemImpropability; // how improbable it is, that an item is released, when a block is destroyed
	
	public ItemController(Model model) {
		this.model = model;
		ItemType.init();
		itemImpropability = defaultItemImpropability;
	}
	
	/**
	 * Capture the item and call the adequade methods.
	 */
	public void capture(Item item) {
		Pad pad = model.getPad();
		Ball[] balls = model.getBalls();
		if (item.getType()!=ItemType.ORANGE) setBallsInvincible(false);
		if (item.getType()!=ItemType.GREEN) unbindBalls();
		
		if (item.getType()!=ItemType.RED) 
			model.clearShots();

		if (item.getType()!=ItemType.BLACK) 
			model.addScore(10);
		if (item.getType()==ItemType.YELLOW) {
			pad.expandWidth();
		
		} else if (item.getType()==ItemType.ORANGE) {
			setBallsInvincible(true);
		
		} else if (item.getType()==ItemType.RED) {
			model.increaseShots();
		
		} else if (item.getType()==ItemType.PURPLE) {
			pad.reduceWidth();
		
		} else if (item.getType()==ItemType.BLUE) {
			splitBall(balls,1);
			
		} else if (item.getType()==ItemType.GREEN) {
			for (Ball ball:balls) 
				ball.bind = true;
		
		} else if (item.getType()==ItemType.BLACK) {
			capture(createRandomItem(item.getX(),item.getY()));
		
		} else if (item.getType()==ItemType.WHITE) {
			model.increaseLives();
		
		} else {
			throw new IllegalArgumentException("Such an item can't exist: type="+item.getType());
		}
		item.setCatched();
	}
	
	public void update(Observable o, Object arg) {
		BallEvent event = (BallEvent) arg;
		if (event.getType()==TYPE.BLOCK_DEATH) { // release a new item
			if (itemImpropability!=0 && new Random().nextInt(itemImpropability)!=0) return;
			int x = event.getX();
			int y = event.getY();
			Item item = createRandomItem(x, y);
			model.addItem(item);
		} else if (event.getType()==TYPE.ITEM_RELEASED) { // a block has released a specific item
			if (event instanceof ItemReleasedEvent) {
				ItemType itemType = ((ItemReleasedEvent) event).getItemType();
				Item item = new Item(event.getX(), event.getY(), itemType);
				model.addItem(item);
			} else {
				throw new IllegalArgumentException("a BallEvent with type ITEM_RELEASED must be a ItemReleasedEvent");
			}
		} else if (event.getType()==TYPE.RESTERT){
			setItemImprobability(defaultItemImpropability);
		}
	}
	
	public void setItemImprobability(int improp) {
		itemImpropability = improp;
	}
	
	/**
	 * Split every ball of the array balls into times balls
	 */
	private void splitBall(final Ball[] balls, final int times) {
		int count = model.getBallCount();
		for (Ball ball:balls) {
			if (count>=ballMax)return;
			double angle = Math.atan2(ball.vy,ball.vx);
			for (int i=0;i<times && count<=ballMax;i++) {
				Ball ball1 = new Ball(ball);
				Ball ball2 = new Ball(ball);
				ball1.vx = (float) Math.cos(angle-(i+1)*0.1f)*ball1.velocity;
				ball1.vy = (float) Math.sin(angle-(i+1)*0.1f)*ball1.velocity;
				ball2.vx = (float) Math.cos(angle+(i+1)*0.1f)*ball2.velocity;
				ball2.vy = (float) Math.sin(angle+(i+1)*0.1f)*ball2.velocity;
				model.addBall(ball1);
				model.addBall(ball2);
				count += 2;
			}
		}
	}
	
	/**
	 * Set all balls invincible (golden) or not (iron)
	 */
	private void setBallsInvincible(boolean b) {
		for (Ball ball:model.getBalls())
			ball.setInvincible(b);
	}
	
	/**
	 * Release all balls
	 */
	private void unbindBalls() {
		for (Ball ball:model.getBalls()) {
			ball.bind = false;
		}
		model.getPad().releaseBoundedBalls();
	}
	
	/**
	 * Create a item of random type at (x,y)
	 */
	private Item createRandomItem(int x, int y) {
		ItemType itemImage = ItemType.getRandomItemType();
		Item item = new Item(x,y,itemImage);
		return item;
	}
	
	/**
	 * Execute the most powerful cheat of Vajanoid
	 */
	public void lantersoftCheat() {
		unbindBalls();
		splitBall(model.getBalls(),15);
		setBallsInvincible(true);
	}
}
