package ch.lanter.vajanoid;

import java.util.ArrayList;
import java.util.List;
import java.util.Observable;

import ch.lanter.vajanoid.BallEvent.TYPE;
import ch.lanter.vajanoid.level.Level;
import ch.lanter.vajanoid.level.LevelChain;

/**
 * The Model contains the state of the game. The levels, the balls, the shots,
 * the items and all sort of data are stored in here. The Model extends
 * Observable so any Observer can listen to BallEvents the Model might launch.
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class Model extends Observable {

	/**
	 * For that blocks know their defaulte boundaries without passing the model
	 * through all the classes down to them, these values are static
	 */
	public static final int WIDTH = 384;
	public static final int HEIGHT = 340;
	
	private List<Ball> balls; // the balls
	private Pad pad; // the pad
	private Level level; // current level
	
	private LevelChain levelChain; // current level chain
	
	private ArrayList<Item> items = new ArrayList<Item>();
	private ArrayList<Shot> shots = new ArrayList<Shot>();
	
	private long leveltime; // how long the current level has been played
	private int width, height;
	private int lives;
	private int score;
	private int shotCount;
	private boolean showCursor; // true, if the cursor should be shown
	private boolean cheated;
	
	public Model() {
		width = WIDTH;
		height = HEIGHT;
		showCursor = true;
		cheated = false;
		lives = 2;
		levelChain = new LevelChain(this);
		balls = new ArrayList<Ball>();
		reset();
	}
	
	/**
	 * Restart the game
	 */
	public void restart() {
		levelChain = new LevelChain(this);
		cheated = false;
		lives = 2;
		score = 0;
		notifyObservers(new BallEvent(-1,-1,TYPE.RESTERT));
		reset();
	}
	
	/**
	 * Reset the model before every new ball
	 */
	public void reset() {
		balls.clear();
		Ball ball = new Ball(this);
		balls.add(ball);
		if (pad!=null) pad = new Pad(pad);
		else { pad = new Pad(width);
			pad.setPosY(height-40);
			pad.setPosX(width/2-pad.getPadWidth()/2);
		}
		ball.x = pad.getPosX()+pad.getPadWidth()/2-ball.width/2+1;
		ball.y = pad.getPosY()-ball.height+3;
		shotCount = 0;
		pad.bindBall(ball);
		items.clear();
		shots.clear();
	}
	
	/**
	 * Update the Ball-, the Item- and the Shot objects regarding the elapsed time 'dt'
	 * @param dt - time elapsed
	 */
	public void step(long dt) {
		leveltime += dt;
		for (Ball ball:balls) ball.step(dt);
		for (Item item:items) item.step(dt);
		for (Shot shot:shots) shot.step(dt);
		level.step(dt);
	}
	
	/**
	 * Let other objects fire Events
	 */
	@Override
	public void notifyObservers(Object arg) {
		setChanged();
		super.notifyObservers(arg);
	}
	
	// Item methods
	public void addItem(Item item) { // delegate
		items.add(item);
	}
	
	public void removeItem(Item item) { // delegate
		items.remove(item);
	}
	
	public Item[] getItems() {
		Item[] ret = new Item[items.size()];
		return items.toArray(ret);
	}
	
	// Ball methods
	public Ball[] getBalls() {
		Ball[] ret = new Ball[balls.size()];
		return balls.toArray(ret);
	}
	
	public void addBall(Ball ball) { // delegate
		balls.add(ball);
	}
	
	public void removeBall(Ball ball) { // delegate
		balls.remove(ball);
	}
	
	public int getBallCount() {
		return balls.size();
	}
	
	// Pad methods
	public Pad getPad() {
		return pad;
	}
	
	// Level methods
	/**
	 * return current level
	 */
	public Level getLevel() {
		return level;
	}
	
	public void nextLevel(int jump) {
		if (jump<=0 || !levelChain.hasNext()) return;
		if (level!=null) level.stop();
		this.deleteObserver(level);
		this.level = levelChain.getNext(jump);
		this.addObserver(level);
		this.level.start();
		leveltime = 0;
		notifyObservers(new BallEvent(-1,-1,TYPE.NEXT_LEVEL));
	}
	
	public boolean hasNextLevel() { // delegate
		return levelChain.hasNext();
	}
	
	public int getJumpsToLevel(String str) { // delegate
		return levelChain.getJumpsToLevel(str);
	}
	
	public int getLevelNumber() { // delegate
		return levelChain.getNumber()+1;
	}
	
	// Life methods
	public void decreaseLives() {
		lives--;
		notifyObservers(new BallEvent(-1,-1,TYPE.LIFE_LOST));
	}
	
	public void increaseLives() {
		lives++;
		notifyObservers(new BallEvent(-1,-1,TYPE.LIFE_GAINED));
	}

	public int getLives() {
		return lives;
	}
	
	// Model methods
	public int getWidth() {
		return width;
	}
	
	public int getHeight() {
		return height;
	}
	
	public boolean isShowCursor() {
		return showCursor;
	}
	
	public void setShowCursor(boolean b) {
		this.showCursor = b;
		notifyObservers(new BallEvent(-1,-1,TYPE.CURSOR));
	}
	
	// Shot methods
	public void increaseShots() {
		shotCount++;
		pad.setShots(shotCount);
	}
	
	public void clearShots() {
		shotCount = 0;
		pad.setShots(shotCount);
	}
	
	public int getShotCount() {
		return shotCount;
	}
	
	public void addShot(Shot shot) {
		shots.add(shot);
	}
	
	public Shot[] getShots() {
		Shot[] ret = new Shot[shots.size()];
		return shots.toArray(ret);
	}
	
	public void removeShot(Shot shot) {
		shots.remove(shot);
	}
	
	// Anti-Cheat methods (maybe not that useful anymore, since  i provide the source code)
	public void setCheated(boolean b) {
		this.cheated = b;
		notifyObservers(new BallEvent(-1,-1,TYPE.CHEATED));
	}
	
	public boolean hasCheated() {
		return cheated;
	}
	
	
	// score methods
	public void addScore(int add) {
		long max = 3*60*1000;
		if (leveltime<max) {
			long plus = add*(max-leveltime)/(max)+1;
			score += plus;
			notifyObservers(new BallEvent(-1,-1,TYPE.SCORED));
		}
	}
	
	public int getScore() {
		return score;
	}
}
