package ch.lanter.vajanoid.block;

import java.awt.AlphaComposite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;

import ch.lanter.vajanoid.Ball;
import ch.lanter.vajanoid.Model;
import ch.lanter.vajanoid.Positionable;
import ch.lanter.vajanoid.Shot;

/**
 * Block defines the look of a block. Furthermore it contains a behavior,
 * describing what is going to happen, when a ball, shot or further object
 * touches the block or even destroys it. Every block is to contain an instance
 * of such a behavior class which is to be called every time, when there has
 * been a contact with a ball, shot or further object, according to the state,
 * whether the block now is destroyed or not.
 * 
 * @author Martin Lanter, Lantersoft 2010
 */
public abstract class Block implements Positionable {

	private static final long serialVersionUID = 201001130153L;

	private boolean rock; // true, if a normal ball can't destroy this block. Only a golden (invicible) ball is able to destroy it

	private boolean destroyed; // true, if this block has been destroyed. Destroyed Block musn't be visible
	private boolean visible;

	private int x,y; // component's topleft location to paint on graphics
	private int w,h; // component's width and height (if the block is split into pieces w,h are the whole width and height, where the block is to paint
	private int xx,yy; // real topleft position of the block (may be out of the view)
	private int ww,hh; // width and height that 
	private int xmin, ymin, xmax, ymax;
	private float opacity;

	private Behavior behavior; // What happens, when a ball or shot hits the block

	public Block(int xx, int yy, int ww, int hh) {
		this.xx = xx;
		this.yy = yy;
		this.rock = false;
		this.destroyed = false;
		this.xmin = 0;
		this.ymin = 0;
		this.xmax = Model.WIDTH;
		this.ymax = Model.HEIGHT;
		this.opacity = 0f;
		this.visible = true;
		setBehavior(null); // must be declared by the subclasses
		setBounds(xx,yy,ww,hh);
	}

	abstract public Image getImage();
	abstract public void interract(Ball ball);
	abstract public void interract(Shot shot);

	public boolean intersects(Positionable pos) {
		return (isVisible() && getBounds().intersects(pos.getBounds()) );
	}

	/**
	 * Save xx and yy and than call setBounds, which evaluates x and y
	 */
	public void setLocation(int xx, int yy) {
		this.xx = xx;
		this.yy = yy;
		setBounds(xx,yy,ww,hh);
	}

	/**
	 * 
	 * @param xx - 
	 * @param yy - 
	 * @param ww - 
	 * @param hh - 
	 */
	private void setBounds(int xx, int yy, int ww, int hh) {
		int diffx = xmax-xmin;
		int diffy = ymax-ymin;
		this.x = this.xx = xx = xmin+((xx-xmin)%diffx+diffx)%diffx;
		this.y = this.yy = yy = ymin+((yy-ymin)%diffy+diffy)%diffy;
		this.w = this.ww = ww;
		this.h = this.hh = hh;

		/*
		 * if the block exceeds the maximum it must be split into pieces. That
		 * is, setting x to xmin and w to full length of diffx. That means, the
		 * leftmost pixel has x coordinates xmin and the rightmost pixel x+w
		 * (=xmin+diffx). Same for yy and hh.
		 */
		if (xx+ww>xmax) {
			this.x = xmin;
			this.w = diffx;
		}
		if (yy+hh>ymax) {
			this.y = ymin;
			this.h = diffy;
		}
	}

	public void paint(Graphics g) {
		if (!isVisible() || opacity<=0) return;
		else if (opacity<1) { // paint transparent
			Graphics2D g2 = (Graphics2D) g.create();
			g2.setComposite(AlphaComposite.SrcOver.derive(opacity));
			g = g2;
		}

		Image image = getImage();
		int wImg = image.getWidth(null);
		int hImg = image.getHeight(null);

		/*
		 * Choose how to paint, depending on whether the block is vertically or
		 * horizontally split up respectively
		 */
		if (xx+ww>xmax && yy+hh>ymax) { // split into 4 pieces
			g.drawImage(image,xmin,ymin,xmin+xx+ww-xmax,ymin+yy+hh-ymax, wImg+xmax-xx-ww,hImg+ymax-yy-hh,wImg,hImg,null); // top left
			g.drawImage(image,xx,ymin,xmax,ymin+yy+hh-ymax, 0,hImg+ymax-yy-hh,xmax-xx,hImg,null); // top right
			g.drawImage(image,xmin,yy,xmin+xx+ww-xmax,ymax, wImg+xmax-xx-ww,0,wImg,ymax-yy,null); // bottom left
			g.drawImage(image,xx,yy,xmax,ymax, 0,0,xmax-xx,ymax-yy,null); // bottom right

		} else if (xx+ww>xmax) { // horizontal split
			g.drawImage(image,xmin,yy,xmin+xx+ww-xmax,yy+hh, wImg+xmax-xx-ww,0,wImg,hImg,null); // left
			g.drawImage(image,xx,yy,xmax,yy+hh, 0,0,xmax-xx,hImg,null); // rigth

		} else if (yy+hh>ymax) { // vertical split
			g.drawImage(image,xx,ymin,xx+ww,ymin+yy+hh-ymax, 0,hImg+ymax-yy-hh,wImg,hImg,null);
			g.drawImage(image,xx,yy,xx+ww,ymax, 0,0,wImg,ymax-yy,null);

		} else { // block is not split up
			g.drawImage(image,xx,yy,ww,hh,null);
		}
	}

	/**
	 * Used for disappearing and appearing blocks. A destroyed block cannot be
	 * visible. If necessary the caller first must call setDestroyed(false).
	 */
	public void setActive(boolean activ) {
		setVisible(!destroyed && activ);
	}

	/**
	 * Returns, whether the block is still to be destoryed or not. Rock blocks
	 * are not alive by default. Likewise are destoyed block not alive.
	 */
	public boolean isAlive() {
		return !rock && !destroyed;
	}

	public int getBlockX() {
		return xx;
	}

	public void setBlockX(int xx) {
		setLocation(xx,yy);
	}

	public int getBlockY() {
		return yy;
	}

	public void setBlockY(int yy) {
		setLocation(xx,yy);
	}

	public int getBlockWidth() {
		return ww;
	}

	public void setBlockWidth(int ww) {
		setBounds(xx,yy,ww,hh);
	}

	public int getBlockHeight() {
		return hh;
	}

	public void setBlockHeight(int hh) {
		setBounds(xx,yy,ww,hh);
	}

	public boolean isRock() {
		return rock;
	}

	public void setRock(boolean b) {
		this.rock = b;
	}

	public int getXMin() {
		return xmin;
	}

	public void setXMin(int xxmin) {
		this.xmin = xxmin;
		setBounds(xx,yy,ww,hh);
	}

	public int getXMax() {
		return xmax;
	}

	public void setXMax(int xxmax) {
		this.xmax = xxmax;
		setBounds(xx,yy,ww,hh);
	}

	public int getYMin() {
		return ymin;
	}

	public void setYMin(int ymin) {
		this.ymin = ymin;
		setBounds(xx,yy,ww,hh);
	}

	public int getYMax() {
		return ymax;
	}

	public void setYMax(int ymax) {
		this.ymax = ymax;
		setBounds(xx,yy,ww,hh);
	}

	public float getOpacity() {
		return opacity;
	}

	public void setOpacity(float opacity) {
		this.opacity = opacity>1?1:opacity<0?0:opacity;
	}

	public boolean isDestroyed() {
		return destroyed;
	}

	/**
	 * setDestroyed(true) automaticly calls setVisible(false),
	 * setDestroyed(false) however doesn't
	 */
	public void setDestroyed(boolean destroyed) {
		this.destroyed = destroyed;
		if (destroyed) setVisible(false);
	}

	public Behavior getBehavior() {
		return behavior;
	}

	public void setBehavior(Behavior behavior) {
		this.behavior = behavior;
	}

	public void setVisible(boolean b) {
		this.visible = b;
	}

	public boolean isVisible() {
		return visible;
	}

	public Rectangle getBounds() {
		return new Rectangle(x,y,w,h);
	}

	public int getX() {
		return x;
	}

	public int getY() {
		return y;
	}

	public int getWidth() {
		return w;
	}

	public int getHeight() {
		return h;
	}
}
