package ch.lanter.vajanoid.block;

import ch.lanter.vajanoid.Eventable;
import ch.lanter.vajanoid.ItemReleasedEvent;
import ch.lanter.vajanoid.ItemType;

/**
 * The ItemerBahavior always releases an item, when it's hit by a ball and then
 * deactivates itself for some time.
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class ItemerBehavior implements Behavior {

	public static final int DEFAULT_DELAY = 150; 
	
	private boolean available; // true if a ball can hit the block and release an item
	private int delay; // time to stay deactivated if a ball hits the block
	
	private ItemType itemType; // null implies a random choice
	private Block block; // the block the behavior belongs to
	
	/**
	 * @param block - the block the behavior belongs to
	 * @param itemType - the itemtype the block is to release when hit
	 */
	public ItemerBehavior(Block block, ItemType itemType) {
		this(block,itemType,DEFAULT_DELAY);
	}
	
	/**
	 * @param block - the block the behavior belongs to
	 * @param itemType - the itemtype the block is to release when hit
	 * @param delay - initial delay
	 */
	public ItemerBehavior(Block block, ItemType itemType, int delay) {
		this.block = block;
		this.itemType = itemType;
		this.delay = delay;
		this.available = true;
	}
	
	@Override
	public void blockStroked(int x, int y, Eventable eventable) {
		if ( !available) return;
		launchItem(x,y,eventable); // only for balls
	}
	
	public void blockShot(int x, int y, Eventable eventable) { }
	
	@Override
	public void blockDestroyed(int x, int y, int hp, Eventable eventable) {
		if ( !available) return;
		eventable.addScore(10*hp);
		launchItem(x,y,eventable); // for shots and balls
	}
	
	private void launchItem(int x, int y, Eventable eventable) {
		if (itemType==null) { 
			eventable.launchEvent(new ItemReleasedEvent(x,y,ItemType.BLACK));
		} else {
			eventable.launchEvent(new ItemReleasedEvent(x,y,itemType));
		}
		if (!block.isDestroyed())
			launchDelay();
	}

	/**
	 * Deactivate the block for some time. The delay increases by 1.5 every time
	 * the block has been hit
	 */
	private void launchDelay() {
		available = false;
		final int toDelay = delay;
		this.delay = delay*15/10;
		new Thread() {
			public void run() {
				try {
					block.setOpacity(.5f);
					Thread.sleep(toDelay);
					available = true;
					block.setOpacity(1f);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}.start();
	}
}
