package ch.lanter.vajanoid.gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.Observable;
import java.util.Observer;

import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;

import ch.lanter.vajanoid.Ball;
import ch.lanter.vajanoid.BallEvent;
import ch.lanter.vajanoid.Controller;
import ch.lanter.vajanoid.Item;
import ch.lanter.vajanoid.Model;
import ch.lanter.vajanoid.Shot;
import ch.lanter.vajanoid.BallEvent.TYPE;
import ch.lanter.vajanoid.block.Block;

/**
 * The view paints all the blocks, balls, items and shots that are in the game.
 * On this panel the whole action takes place.
 * 
 * @author Martin Lanter, LanterSoft 2010
 */
public class View extends JPanel implements MouseMotionListener {

	private static final long serialVersionUID = 201001130152L;

	private Model model;

	private JPanel pnlInFront; // Panel in the front (e.g. Pause, press F3)
	private boolean showString; // true, if the panel in the frotn should be displayed
	
	private Cursor hiddenCursor; // an invisible cursor 

	public View(Model model) {
		setLayout(null);
		addMouseMotionListener(this);
		this.model = model;
		this.setSize(model.getWidth(), model.getHeight());
		this.setPreferredSize(new Dimension(model.getWidth(),model.getHeight()));
		this.setMaximumSize(new Dimension(model.getWidth(),model.getHeight()));
		this.setMinimumSize(new Dimension(model.getWidth(),model.getHeight()));
		hiddenCursor = Toolkit.getDefaultToolkit().createCustomCursor(
				new BufferedImage(1,1,BufferedImage.TYPE_4BYTE_ABGR),
				new Point(0,0),
				"Hidden Cursor");
		setAccurateCursor();
		model.addObserver(new Observer() {
			public void update(Observable o, Object arg) {
				BallEvent event = (BallEvent) arg;
				if (event.getType()==TYPE.CURSOR) {
					setAccurateCursor();
				}
			}
		});
	}
	
	/**
	 * Set the default Cursor (Arrow) or the invisible hiddenCursor
	 */
	private void setAccurateCursor() {
		if (model.isShowCursor()) {
			setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		} else {
			setCursor(hiddenCursor);
		}
	}

	@Override
	public void paint(Graphics g) {
		long t0 = System.nanoTime();
		super.paint(g);
		long te = System.nanoTime()-t0;
		if (Controller.printWarnings && te>5000000)
			System.err.println("Warning: too long paint time elapsed: "+te/1000+"ys");
	}

	@Override
	public void paintComponent(Graphics g) {
		Graphics2D g2 = (Graphics2D) g.create();
		GradientPaint gradint = new GradientPaint(
				new Point(0,0),new Color(170,180,255),
				new Point(0,getHeight()/2),new Color(150,150,255));
		g2.setPaint(gradint);
		Rectangle clip = g.getClipBounds();
		g2.fillRect(clip.x, clip.y, clip.width, clip.height);

		paintObjects(g);

		if (showString) {
			String text = "Martin Lanter, Lantersoft 2010";
			g2.setFont(new Font(Font.SANS_SERIF,Font.BOLD,14));
			FontMetrics metrics = g2.getFontMetrics();
			int x = getWidth()/2-metrics.stringWidth(text)/2;
			int y = getHeight()-10;
			g2.setColor(new Color(255,255,255,100));
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
			g2.drawString(text,x,y);
		}
		g2.dispose();
	}

	/**
	 * Paint the balls, blocks, items, shots and the pad onto Graphics g
	 */
	private void paintObjects(Graphics g) {
		for (Ball b:model.getBalls())
			b.paint(g);
		for (Block b:model.getLevel().getBlocks())
			if (b!=null) b.paint(g);
		for (Item i:model.getItems())
			i.paint(g);
		for (Shot s:model.getShots())
			s.paint(g);
		model.getPad().paint(g);
	}

	/**
	 * Set a textstring into the front
	 * @param strTop - big size text
	 * @param str2 - small size text
	 */
	public void setStringInFront(String strTop, String str2) {
		removeStringInFront();
		showString = true;
		pnlInFront = new JPanel();
		pnlInFront.setLayout(new BoxLayout(pnlInFront,BoxLayout.Y_AXIS));
		pnlInFront.setOpaque(false);

		JLabel lbl1 = new HighlightLabel(strTop,5);
		lbl1.setFont(new Font(Font.SANS_SERIF,Font.BOLD,48));
		lbl1.setAlignmentX(CENTER_ALIGNMENT);

		JLabel lbl2 = new HighlightLabel(str2,2);
		lbl2.setFont(new Font(Font.SANS_SERIF,Font.BOLD,16));
		lbl2.setAlignmentX(CENTER_ALIGNMENT);

		pnlInFront.add(lbl1);
		pnlInFront.add(lbl2);

		Dimension dim = getSize();
		Dimension prf = pnlInFront.getPreferredSize();
		pnlInFront.setBounds((dim.width-prf.width)/2,(dim.height-prf.height)/2,prf.width,prf.height);
		add(pnlInFront,0); //PALETTE_LAYER

		revalidate();
		repaint();
	}

	/**
	 * Remove the panel in the front if there is any
	 */
	public void removeStringInFront() {
		showString = false;
		if (pnlInFront!=null) {
			remove(pnlInFront);
		}
	}

	/**
	 * Remove all components except for the ball and the pad
	 */
	public void unload() {
		showString = false;
		removeAll(); // StrInFront and Ranking
		repaint();
	}

    @Override public void mouseMoved(MouseEvent e) {
    	model.getPad().moveTo(e.getX());
    }

    @Override public void mouseDragged(MouseEvent e) {
    	model.getPad().moveTo(e.getX());
    }
}
