package ch.lanter.vajanoid.level;

import java.util.Observable;
import java.util.Random;

import org.jdesktop.animation.timing.Animator;
import org.jdesktop.animation.timing.TimingTargetAdapter;

import ch.lanter.vajanoid.block.Block;
import ch.lanter.vajanoid.block.Path;

/**
 * AbstractLevel is a default implementation of the Level interface, containing
 * the basic methods and providing some functionality like executing the level's
 * paths.
 * 
 * @author Martin Lanter, Lantersoft 2010
 */
public abstract class AbstractLevel implements Level {

	@Override
	public abstract Block[] getBlocks();

	@Override
	public boolean isWon() {
		for (Block b:getBlocks()) {
			if (b!=null && b.isAlive()) { // there is any block still alive
				return false;
			}
		}
		return true;
	}

	@Override
	public void start() {
		final Block[] blocks = getBlocks();
		final int n = blocks.length;
		final float[] offset = new float[n];
		Random rand = new Random();
		for (int i=0;i<n;i++) {
			offset[i] = rand.nextFloat();
		}
		final int duration = 2000;
		Animator animator = new Animator(duration,new TimingTargetAdapter() {
			 public void timingEvent(float fraction) {
				 for (int i=0;i<n;i++) {
					 float opacity = -offset[i]+2*fraction;
					 if (blocks[i]!=null)
						 blocks[i].setOpacity(opacity>=0?opacity:0);
				 }
			 }
			 public void end() {
				 for (int i=0;i<n;i++) {
					 if (blocks[i]!=null)
						 blocks[i].setOpacity(1f);
				 }
			 }
		});
		animator.setResolution(50);
		animator.start();
	}
	
	@Override
	public void update(Observable o, Object arg) { }
	
	/**
	 * If there are paths in a subclass they should be returned by overriding
	 * this method. AbstractLevel than performs automaticly the step(dt) call.
	 */
	public Path[] getPaths() {
		return null;
	}
	
	/**
	 * Let all paths perform a step with the given elapsed time dt
	 */
	@Override
	public void step(long dt) {
		Path[] paths = getPaths();
		if (paths==null) return;
		for (Path path:paths)
			if (path!=null)
				path.step(dt);
	}
	
	// These methods aren't necessarily doing anything
	@Override
	public void pause() { }
	
	@Override
	public void resume() { }
	
	@Override
	public void stop() { }
	
}
