package ch.lanter.vajanoid.path;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;

import ch.lanter.vajanoid.block.Block;

/**
 * PolygonalPath arranges its blocks on a polygonal shape and lets them slide
 * along the boundaries.
 * 
 * @author Martin Lanter, LanterSoft
 */
public class PolygonalPath extends AbstractPath {

	private float s; // 
	private float v; // speed in pixel per second
	
	private ArrayList<Point> vertices;
	private float[] dists; // dist[n] is the total way around the polygon
	
	public PolygonalPath(float v) {
		this.s = 0;
		this.v = v;
		this.vertices = new ArrayList<Point>();
	}
	
	public void addVertice(Point p) {
		vertices.add(p);
	}
	
	public void addEmptyBlock() {
		super.addBlock(null);
	}
	
	/**
	 * Arrange the blocks uniformly along the boundaries
	 */
	public void arrange() {
		int n = vertices.size();
		if (n==0) throw new IllegalStateException("PolygonalPath has no Vertices");
		dists = new float[n+1];
		for (int i=0;i<n;i++) {
			int dx = vertices.get(i).x-vertices.get((i+1)%n).x;
			int dy = vertices.get(i).y-vertices.get((i+1)%n).y;
			// (dy*1.6)^2 because the blocks are 1.6 times wide as heigh. This makes the movement smoothly
			dists[i+1] = dists[i] + (float)Math.sqrt(dx*dx + dy*dy*1.6*1.6);
		}
		int m = getCount();
		int ptr = 0;
		float dt = dists[n]/m;
		Iterator<Block> iterator = getBlockIterator();
		for (int i=0;i<m;i++) {
			Block block = iterator.next();
			while(dists[ptr+1]<i*dt) {
				ptr++;
			}
			float u = (i*dt-dists[ptr])/(dists[ptr+1]-dists[ptr]); // way from ptr to block location
			float vx = (vertices.get((ptr+1)%n).x-vertices.get(ptr).x);
			float vy = (vertices.get((ptr+1)%n).y-vertices.get(ptr).y);
			if (block!=null) {
				block.setBlockX(Math.round(vertices.get(ptr).x+u*vx));
				block.setBlockY(Math.round(vertices.get(ptr).y+u*vy));
			}
		}
	}

	/**
	 * Forward all the blocks according to the passed time dt. dt is the time
	 * passed since the last call of step
	 */
	@Override
	public void step(long dt) {
		float ds = dt*v/1000;
		s = s+ds;
		
		int n = vertices.size();
		int m = getCount();
		int ptr = 0;
		float d = dists[n]/m;

		s = (s%dists[n]+dists[n])%dists[n];
		
		Iterator<Block> iterator = getBlockIterator();
		for (int i=0;i<m;i++) {
			Block block = iterator.next();
			while( ( (i*d+s)%dists[n]-dists[ptr] + dists[n])%dists[n]> dists[ptr+1]-dists[ptr]) {
				ptr = (ptr+1)%n;
			}
			float dist = (dists[ptr+1]-dists[ptr]);
			float u = ((i*d+s)-dists[ptr])%dists[n] /dist; // way from ptr to block location
			float vx = (vertices.get((ptr+1)%n).x-vertices.get(ptr).x);
			float vy = (vertices.get((ptr+1)%n).y-vertices.get(ptr).y);
			if (block!=null) {
				block.setBlockX(Math.round(vertices.get(ptr).x+u*vx));
				block.setBlockY(Math.round(vertices.get(ptr).y+u*vy));
			}
		}
	}
	
	public int getVerticesCount() {
		return vertices.size();
	}
}
